#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "showtext.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

ShowText::ShowText(int w, int h) {
	width = w;
	height = h;
	font = NULL;
	fontChars = NULL;
	lowerCase = false;
	charWidth = charHeight = 0;
	fontImageWidth = 0;
	plain = true;
	color = 0x00FFFFFF;
	texts = NULL;
	textNum = 0;
	bufferWidth = bufferHeight = 0;
	textTime = startTime = 0;
}

ShowText::~ShowText(void) {
	if (texts==NULL) return;
	unsigned int **ptr = texts;
	unsigned int *b;
	while ((b=*ptr++)!=NULL)
		free(b);
	free(texts);
}

void ShowText::init(void) {
}

void ShowText::setFont(unsigned int *f, char *chars, bool lc, int cw, int ch, int w, bool p, unsigned int c) {
	font = f;
	fontChars = chars;
	lowerCase = lc;
	charWidth = cw;
	charHeight = ch;
	fontImageWidth = w;
	plain = p;
	color = c;
}

void ShowText::setText(char **text, int time) {
	int maxChars = 0;
	int maxRows = 0;

	char **ptr = text;
	char *s;
	int chars, rows;
	int n = 0;
	while ((s=*ptr++)!=NULL) {
		n++;
		chars = 0;
		rows = 1;

		while (*s!='\0') {
			if (*s=='\n') {
				rows++;
				if (chars>maxChars) maxChars = chars;
				chars = 0;
			} else {
				chars++;
			}
			s++;
		}
		if (chars>maxChars) maxChars = chars;
		if (rows>maxRows) maxRows = rows;
	}

	bufferWidth = maxChars*charWidth;
	bufferHeight = maxRows*charHeight;

	texts = (unsigned int**) malloc(sizeof(int*)*(n+1));
	ptr = text;
	int i = 0;
	while ((s=*ptr++)!=NULL) {
		texts[i] = (unsigned int*) malloc(sizeof(int)*bufferWidth*bufferHeight);
		memset(texts[i],0,sizeof(int)*bufferWidth*bufferHeight);
		drawString(texts[i],s);
		i++;
	}
	texts[i] = NULL;
	textNum = n;

	textTime = time/n;

	startTime = -1;
}

void ShowText::drawString(unsigned int *buffer, char *string) {
	int rows = 1;
	char *ptr = string;
	while (*ptr!='\0') {
		if (*ptr=='\n') rows++;
		ptr++;
	}

	int y = (bufferHeight-rows*charHeight)/2;

	char *tempstring = (char*) malloc(strlen(string)+1);
	ptr = string;
	char *outptr = tempstring;
	while (*ptr!='\0') {
		if (*ptr=='\n') {
			*outptr = '\0';
			drawRow(buffer,tempstring,y);
			outptr = tempstring;
			y += charHeight;
		} else {
			*outptr++ = *ptr;
		}
		ptr++;
	}
	*outptr = '\0';
	drawRow(buffer,tempstring,y);
	free(tempstring);
}

void ShowText::drawRow(unsigned int *buffer, char *row, int y) {
	int length = strlen(row);
	int x = (bufferWidth-length*charWidth)/2;
	for (int i = 0; i < length; i++) {
		drawChar(buffer,*row++,x,y);
		x += charWidth;
	}
}

int ShowText::getIndex(char c) {
	if (!lowerCase) c = toupper(c);
	char *pos = strchr(fontChars,c);
	if (pos==NULL) return -1;
	return (pos-fontChars)/sizeof(char);
}

void ShowText::drawChar(unsigned int *buffer, char c, int bufferX, int bufferY) {
	int index = getIndex(c);

	int fontX = charWidth*(index%fontImageWidth);
	int fontY = charHeight*(index/fontImageWidth);

	int fontColor;
	unsigned int *outptr;
	unsigned int *inptr;
	for (int y = 0; y < charHeight; y++) {
		outptr = buffer + (bufferY+y)*bufferWidth+bufferX;
		inptr = font + (fontY+y)*fontImageWidth*charWidth+fontX;
		for (int x = 0; x < charWidth; x++) {
			fontColor = (*inptr++)&0x00FFFFFF;
			if (fontColor)
				*outptr++ = 0xFF000000 | fontColor;
			else
				*outptr++ = 0x00000000;
		}
	}
}

void ShowText::draw(unsigned int *buffer, int time) {
	if (startTime<0) startTime = time;
	time -= startTime;

	int textNr = time/textTime;
	int textPos = time%textTime;

	if (textNr>=textNum) return;

	int x = (width-bufferWidth)/2;
	int y;

	if (textPos < 20*textTime/100) {
		// Fall
		y = (int) ((height-bufferHeight)/2 - ((height-bufferHeight)/2 + bufferHeight)*cos((M_PI/2)*textPos/(20*textTime/100)));
	} else if (textPos < 30*textTime/100) {
		// Bounce
		y = (int) ((height-bufferHeight)/2 - (0.20*(height-bufferHeight)/2)*sin(M_PI*(textPos-20*textTime/100)/(10*textTime/100)));
	} else if (textPos < 75*textTime/100) {
		y = (height-bufferHeight)/2;
	} else {
		// Fall
		y = (int) (height - (height-(height-bufferHeight)/2)*cos((M_PI/2)*(textPos-75*textTime/100)/(25*textTime/100)));
	}

	drawOverlay(buffer,texts[textNr],x,y);
}

void ShowText::drawOverlay(unsigned int *buffer, unsigned int *overlay, int x, int y) {
	int overlayXStart = 0, overlayYStart = 0;
	int bufferXStart = x, bufferYStart = y;

	if (x<0) {
		overlayXStart = -x;
		bufferXStart = 0;
	}
	if (y<0) {
		overlayYStart = -y;
		bufferYStart = 0;
	}

	int bx, by;
	by = bufferYStart;
	int c;
	unsigned int *outptr, *inptr;
	for (int oy = overlayYStart; oy < bufferHeight; oy++) {
		if (by>=height) break;
		bx = bufferXStart;
		outptr = buffer + by*width + bufferXStart;
		inptr = overlay + oy*bufferWidth + overlayXStart;
		for (int ox = overlayXStart; ox < bufferWidth; ox++) {
			if (bx>=width) break;
			c = (*inptr++)&0x00FFFFFF;
			if (c)
				if (plain)
					*outptr = color;
				else
					*outptr = c;
			outptr++;
			bx++;
		}
		by++;
	}
}

